﻿
#include <tchar.h>

#include <locale.h>

#include <stdio.h>

#include <Windows.h>

#include "Dumpper.hpp"


enum RV : int
{
	SUCCESS = 0,

	FAIL,

	INVALID_ARG,
	FAIL_OPEN,
	FAIL_SEEK,
	FAIL_READ,
	FAIL_WRITE,
	FAIL_SIZE,
};


class CDumper
{

protected:

	BYTE* m_lpBuffer;

	
	INT_PTR (CDumper::* m_lpfnDisplay)( BYTE const* lpBuffer, LARGE_INTEGER* lpRemaningLengt );


public:

	LARGE_INTEGER m_offset;


protected:

	INT_PTR DisplayFirst( BYTE const* lpBuffer, LARGE_INTEGER* lpRemaningLength );

	INT_PTR DisplayNext( BYTE const* lpBuffer, LARGE_INTEGER* lpRemaningLength );


public:

	CDumper();

	void BeginDisplay();

	INT_PTR Display( BYTE const* lpBuffer, LARGE_INTEGER length );
};

const BYTE ASCII[] =
{
	_T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'),
	_T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'),

	_T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'),
	_T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'), _T('.'),

	_T(' '), _T('!'), _T('"'), _T('#'), _T('$'), _T('%'), _T('&'), _T('\''),
	_T('('), _T(')'), _T('*'), _T('+'), _T(','), _T('-'), _T('.'), _T('/'),

	_T('0'), _T('1'), _T('2'), _T('3'), _T('4'), _T('5'), _T('6'), _T('7'),
	_T('8'), _T('9'), _T(':'), _T(';'), _T('<'), _T('='), _T('>'), _T('?'),

	_T('@'), _T('A'), _T('B'), _T('C'), _T('D'), _T('E'), _T('F'), _T('G'),
	_T('H'), _T('I'), _T('J'), _T('K'), _T('L'), _T('M'), _T('N'), _T('O'),

	_T('P'), _T('Q'), _T('R'), _T('S'), _T('T'), _T('U'), _T('V'), _T('W'),
	_T('X'), _T('Y'), _T('Z'), _T('['), _T('\\'),_T(']'), _T('^'), _T('_'),

	_T('`'), _T('a'), _T('b'), _T('c'), _T('d'), _T('e'), _T('f'), _T('g'),
	_T('h'), _T('i'), _T('j'), _T('k'), _T('l'), _T('m'), _T('n'), _T('o'),

	_T('p'), _T('q'), _T('r'), _T('s'), _T('t'), _T('u'), _T('v'), _T('w'),
	_T('x'), _T('y'), _T('z'), _T('{'), _T('|'), _T('}'), _T('~'), _T('.'),
};

inline BYTE ToChar( INT_PTR nValue )
{
	if( nValue >= sizeof(ASCII) )
	{
		return (BYTE)_T('.');
	}

	return ASCII[nValue];
}

INT_PTR DisplayDump( HANDLE hFile, LARGE_INTEGER* lpOffset, LARGE_INTEGER* lpSize )
{
	// オフセットとサイズを確認

	LARGE_INTEGER fileLength;

	// ファイルサイズを取得
	if( !GetFileSizeEx(hFile, &fileLength) )
	{
		return RV::FAIL_SIZE;
	}

	// オフセットの有効範囲を確認
	if( lpOffset->QuadPart >= fileLength.QuadPart )
	{
		return RV::INVALID_ARG;
	}

	// 読み込み可能サイズ
	fileLength.QuadPart -= lpOffset->QuadPart;

	if( lpSize->QuadPart >= fileLength.QuadPart )
	{
		lpSize->QuadPart = fileLength.QuadPart;

		if( !lpSize->QuadPart )
		{
			return RV::INVALID_ARG;
		}
	}


	// ファイルシーク

	::SetFilePointer( hFile, lpOffset->LowPart, &lpOffset->HighPart, FILE_BEGIN );

	DWORD nError = ::GetLastError();

	if( ERROR_SUCCESS != nError )
	{
		return RV::FAIL_SEEK;
	}


	CDumpper dumpper;


	INT_PTR nResult = RV::SUCCESS;


	// 読み込みバッファに格納しているサイズ
	LARGE_INTEGER bufferLen;

	// 読み込みバッファ
	BYTE nBuffer[8192];


	dumpper.m_offset = *lpOffset;


	// 指定されたサイズ繰り返す
	while( lpSize->QuadPart > 0 )
	{
		// ファイルから読み込むサイズを算出

		bufferLen.QuadPart = lpSize->QuadPart;

		if( bufferLen.QuadPart > sizeof(nBuffer) )
		{
			bufferLen.QuadPart = sizeof(nBuffer);
		}


		// ファイルから読み込む

		DWORD nReadLength;

		BOOL bRead = ::ReadFile( hFile, nBuffer, bufferLen.LowPart, &nReadLength, NULL );

		if( !bRead )
		{
			nResult = RV::FAIL_SEEK;
			break;
		}


		// 読み込んだデータをダンプ

		LPBYTE lpBuffer = nBuffer;

		while( nReadLength > 0 )
		{
			INT_PTR nDumpLength = dumpper.Display( lpBuffer, *lpSize );

			lpBuffer += nDumpLength;

			lpSize->QuadPart -= nDumpLength;

			nReadLength -= (DWORD)nDumpLength;
		}
	}


	return nResult;
}

int _tmain( int nArgc, _TCHAR* lpszArgv[] )
{
	_tsetlocale( LC_ALL, _T("Japanese") );

#if 1

	// 引数の個数を確認
	if( nArgc == 1 )
	{
		_TCHAR const szUsage[] =
		{
			_T("指定したファイルをダンプする\n")
			_T("2 GB 以上のファイルサイズに対応\n")
			_T("\n")
			_T("書式\n")
			_T("Dump64 <ファイル・パス> <オフセット> <サイズ>\n")
			_T("\n")
			_T("オフセットとサイズは 8 進数または 16 進数でも指定可\n")
			_T("    32 = 0x20 (16 進数。先頭に 0x を付ける)\n")
			_T("    32 =  040 ( 8 進数。先頭に 0 を付ける)\n")
		};

		_putts( szUsage );

		return RV::INVALID_ARG;
	}


	// 引数の個数を確認
	if( nArgc != 4 )
	{
		return RV::INVALID_ARG;
	}


	_TCHAR* lpszTerminate;


	// 読み込み位置を取得

	LARGE_INTEGER offset = { 0 };

	offset.QuadPart = _tcstoll( lpszArgv[2], &lpszTerminate, 0 );

	if( *lpszTerminate != _T('\0') || offset.QuadPart < 0 )
	{
		return RV::INVALID_ARG;
	}


	// 読み込みサイズを取得

	LARGE_INTEGER size = { 0 };

	size.QuadPart = _tcstoll( lpszArgv[3], &lpszTerminate, 0 );

	if( *lpszTerminate != _T('\0') || size.QuadPart <= 0 )
	{
		return RV::INVALID_ARG;
	}


	// ファイルを開いてダンプ

	HANDLE hFile = ::CreateFile( lpszArgv[1], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL );

	if( INVALID_HANDLE_VALUE == hFile )
	{
		return RV::FAIL_OPEN;
	}

	// ダンプ
	int nResult = (int)DisplayDump( hFile, &offset, &size );

	::CloseHandle( hFile );

#else
	LARGE_INTEGER remainLength = { 0x38 };


	CDumpper dumper;

	dumper.m_offset.QuadPart = 0x35;

	BYTE const* lpBuffer = ASCII + 0x35;

	while( remainLength.QuadPart > 0 )
	{
		INT_PTR nCount = dumper.Display( lpBuffer, remainLength );

		remainLength.QuadPart -= nCount;
		lpBuffer += nCount;
	}
#endif

	return nResult;
}
